@@ -1,55 +1,40 @@
-# Copyright (C) 2020 Unitary Fund
+# Copyright (C) Unitary Foundation
 #
-# This program is free software: you can redistribute it and/or modify
-# it under the terms of the GNU General Public License as published by
-# the Free Software Foundation, either version 3 of the License, or
-# (at your option) any later version.
-#
-# This program is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-# GNU General Public License for more details.
-#
-# You should have received a copy of the GNU General Public License
-# along with this program.  If not, see <https://www.gnu.org/licenses/>.
+# This source code is licensed under the GPL license (v3) found in the
+# LICENSE file in the root directory of this source tree.
 
 """Classes corresponding to different zero-noise extrapolation methods."""
-from abc import ABC, abstractmethod
-from copy import deepcopy
-from typing import (
-    Any,
-    Callable,
-    Dict,
-    List,
-    Optional,
-    Sequence,
-    Tuple,
-    Union,
-)
-import warnings
 
-from matplotlib.figure import Figure
+import warnings
+from abc import ABC, abstractmethod
+from collections.abc import Callable, Sequence
+from copy import deepcopy
+from typing import Any, TypeAlias, cast
+
 import matplotlib.pyplot as plt
 import numpy as np
-from numpy.lib.polynomial import RankWarning
-from scipy.optimize import curve_fit, OptimizeWarning
+import numpy.typing as npt
 from cirq import Circuit
+from matplotlib.figure import Figure
+from numpy.lib.polynomial import RankWarning
+from scipy.optimize import OptimizeWarning, curve_fit
 
-from mitiq import QPROGRAM
-from mitiq.collector import Collector
+from mitiq import QPROGRAM, QuantumResult
+from mitiq.executor import Executor
 from mitiq.interface import accept_any_qprogram_as_input
+from mitiq.observable import Observable
+from mitiq.zne.scaling import fold_gates_at_random
 
-
-ExtrapolationResult = Union[
-    float,  # The zero-noise value.
-    Tuple[
+ExtrapolationResult: TypeAlias = (
+    float  # The zero-noise value.
+    | tuple[
         float,  # The zero-noise value.
-        Optional[float],  # The (estimated) error on the zero-noise value.
-        List[float],  # Optimal parameters found during fitting.
-        Optional[np.ndarray],  # Covariance of fitting parameters.
+        float | None,  # The (estimated) error on the zero-noise value.
+        list[float],  # Optimal parameters found during fitting.
+        npt.NDArray[np.float64] | None,  # Covariance of fitting parameters.
         Callable[[float], float],  # Function that was fit.
-    ],
-]
+    ]
+)
 
 
 class ExtrapolationError(Exception):
@@ -110,8 +95,8 @@ def mitiq_curve_fit(
     ansatz: Callable[..., float],
     scale_factors: Sequence[float],
     exp_values: Sequence[float],
-    init_params: Optional[List[float]] = None,
-) -> Tuple[List[float], np.ndarray]:
+    init_params: list[float] | None = None,
+) -> tuple[list[float], npt.NDArray[np.float64]]:
     """Fits the ansatz to the (scale factor, expectation value) data using
     ``scipy.optimize.curve_fit``, returning the optimal parameters and
     covariance matrix of the parameters.
@@ -143,7 +128,7 @@ def mitiq_curve_fit(
             # replace OptimizeWarning with ExtrapolationWarning
             if warn.category is OptimizeWarning:
                 warn.category = ExtrapolationWarning
-                warn.message = _EXTR_WARN  # type: ignore
+                warn.message = _EXTR_WARN
             # re-raise all warnings
             warnings.warn_explicit(
                 warn.message, warn.category, warn.filename, warn.lineno
@@ -157,8 +142,8 @@ def mitiq_polyfit(
     scale_factors: Sequence[float],
     exp_values: Sequence[float],
     deg: int,
-    weights: Optional[Sequence[float]] = None,
-) -> Tuple[List[float], Optional[np.ndarray]]:
+    weights: Sequence[float] | None = None,
+) -> tuple[list[float], npt.NDArray[np.float64] | None]:
     """Fits the ansatz to the (scale factor, expectation value) data using
     ``numpy.polyfit``, returning the optimal parameters and covariance matrix
     of the parameters.
@@ -192,7 +177,7 @@ def mitiq_polyfit(
         # replace RankWarning with ExtrapolationWarning
         if warn.category is RankWarning:
             warn.category = ExtrapolationWarning
-            warn.message = _EXTR_WARN  # type: ignore
+            warn.message = _EXTR_WARN
         # re-raise all warnings
         warnings.warn_explicit(
             warn.message, warn.category, warn.filename, warn.lineno
@@ -218,43 +203,46 @@ class Factory(ABC):
     """
 
     def __init__(self) -> None:
-        self._instack: List[Dict[str, float]] = []
-        self._outstack: List[float] = []
-        self._opt_params: Optional[List[float]] = None
-        self._params_cov: Optional[np.ndarray] = None
-        self._zne_limit: Optional[float] = None
-        self._zne_error: Optional[float] = None
-        self._zne_curve: Optional[Callable[[float], float]] = None
+        self._instack: list[dict[str, float]] = []
+        self._outstack: list[float] = []
+        self._opt_params: list[float] | None = None
+        self._params_cov: npt.NDArray[np.float64] | None = None
+        self._zne_limit: float | None = None
+        self._zne_error: float | None = None
+        self._zne_curve: Callable[[float], float] | None = None
         self._already_reduced = False
-        self._options: Dict[str, float] = {}
+        self._options: dict[str, float | None] = {}
 
-    def get_scale_factors(self) -> np.ndarray:
-        """Returns the scale factors at which the factory has computed
-        expectation values.
+    def get_scale_factors(self) -> list[float]:
+        """Returns the scale factors that were either passed in to the factory
+        or at which the factory has computed expectation values.
         """
-        return np.array(
-            [params.get("scale_factor") for params in self._instack]
-        )
+        scale_factors = [
+            params.get("scale_factor", 0.0) for params in self._instack
+        ]
+        if not scale_factors and hasattr(self, "_scale_factors"):
+            return [scale_factor for scale_factor in self._scale_factors]
+        return scale_factors
 
-    def get_expectation_values(self) -> np.ndarray:
+    def get_expectation_values(self) -> list[float]:
         """Returns the expectation values computed by the factory."""
-        return np.array(self._outstack)
+        return self._outstack
 
-    def get_optimal_parameters(self) -> np.ndarray:
+    def get_optimal_parameters(self) -> list[float]:
         """Returns the optimal model parameters produced by the extrapolation
         fit.
         """
         if self._opt_params is None:
             raise ValueError(DATA_MISSING_ERR)
-        return np.array(self._opt_params)
+        return self._opt_params
 
-    def get_parameters_covariance(self) -> np.ndarray:
+    def get_parameters_covariance(self) -> npt.NDArray[np.float64]:
         """Returns the covariance matrix of the model parameters produced by
         the extrapolation fit.
         """
         if self._params_cov is None:
             raise ValueError(DATA_MISSING_ERR)
-        return np.array(self._params_cov)
+        return self._params_cov
 
     def get_zero_noise_limit(self) -> float:
         """Returns the last evaluation of the zero-noise limit
@@ -293,8 +281,11 @@ class Factory(ABC):
     def run(
         self,
         qp: QPROGRAM,
-        executor: Callable[..., float],
-        scale_noise: Callable[[QPROGRAM, float], QPROGRAM],
+        executor: Executor | Callable[..., QuantumResult],
+        observable: Observable | None = None,
+        scale_noise: Callable[
+            [QPROGRAM, float], QPROGRAM
+        ] = fold_gates_at_random,  # type: ignore [has-type]
         num_to_average: int = 1,
     ) -> "Factory":
         """Calls the executor function on noise-scaled quantum circuit and
@@ -302,8 +293,13 @@ class Factory(ABC):
 
         Args:
             qp: Quantum circuit to scale noise in.
-            executor: Function which inputs a (list of) quantum circuits and
-                outputs a (list of) expectation values.
+            executor: A ``mitiq.Executor`` or a function which inputs a (list
+                of) quantum circuits and outputs a (list of)
+                ``mitiq.QuantumResult`` s.
+            observable: Observable to compute the expectation value of. If
+                None, the `executor` must return an expectation value.
+                Otherwise, the `QuantumResult` returned by `executor` is used
+                to compute the expectation of the observable.
             scale_noise: Function which inputs a quantum circuit and outputs
                 a noise-scaled quantum circuit.
             num_to_average: Number of times the executor function is called
@@ -313,7 +309,8 @@ class Factory(ABC):
 
     @abstractmethod
     def run_classical(
-        self, scale_factor_to_expectation_value: Callable[..., float],
+        self,
+        scale_factor_to_expectation_value: Callable[..., float],
     ) -> "Factory":
         """Calls the function scale_factor_to_expectation_value at each scale
         factor of the factory, and stores the results.
@@ -326,8 +323,12 @@ class Factory(ABC):
         """
         raise NotImplementedError
 
+    @abstractmethod
+    def reduce(self) -> float:
+        raise NotImplementedError
+
     def push(
-        self, instack_val: Dict[str, float], outstack_val: float
+        self, instack_val: dict[str, float], outstack_val: float
     ) -> "Factory":
         """Appends "instack_val" to "self._instack" and "outstack_val" to
         "self._outstack". Each time a new expectation value is computed this
@@ -380,9 +381,10 @@ class Factory(ABC):
         fig = self.plot_data()
 
         smooth_scale_factors = np.linspace(0, self.get_scale_factors()[-1], 20)
-        smooth_expectations = (self.get_extrapolation_curve())(
-            smooth_scale_factors
-        )
+        smooth_expectations = [
+            self.get_extrapolation_curve()(scale_factor)
+            for scale_factor in smooth_scale_factors
+        ]
         plt.xlim(left=0)
         fig.axes[0].plot(
             smooth_scale_factors,
@@ -414,28 +416,26 @@ class BatchedFactory(Factory, ABC):
 
     Specific (non-adaptive) extrapolation algorithms are derived from this
     class by defining the `reduce` method.
+
+    Args:
+        scale_factors: Sequence of noise scale factors at which expectation
+            values should be measured.
+        shot_list: Optional sequence of integers corresponding to the
+            number of samples taken for each expectation value. If this
+            argument is explicitly passed to the factory, it must have the
+            same length of scale_factors and the executor function must
+            accept "shots" as a valid keyword argument.
+
+    Raises:
+        ValueError: If the number of scale factors is less than 2.
+        TypeError: If shot_list is provided and has any non-integer values.
     """
 
     def __init__(
         self,
         scale_factors: Sequence[float],
-        shot_list: Optional[List[int]] = None,
+        shot_list: list[int] | None = None,
     ) -> None:
-        """Constructs a BatchedFactory.
-
-        Args:
-            scale_factors: Sequence of noise scale factors at which expectation
-                values should be measured.
-            shot_list: Optional sequence of integers corresponding to the
-                number of samples taken for each expectation value. If this
-                argument is explicitly passed to the factory, it must have the
-                same length of scale_factors and the executor function must
-                accept "shots" as a valid keyword argument.
-
-        Raises:
-            ValueError: If the number of scale factors is less than 2.
-            TypeError: If shot_list is provided and has any non-integer values.
-        """
         if len(scale_factors) < 2:
             raise ValueError("At least 2 scale factors are necessary.")
 
@@ -462,7 +462,7 @@ class BatchedFactory(Factory, ABC):
 
     @staticmethod
     @abstractmethod
-    def extrapolate(*args, **kwargs) -> ExtrapolationResult:
+    def extrapolate(*args: Any, **kwargs: Any) -> ExtrapolationResult:
         """Returns the extrapolation to the zero-noise limit."""
         raise NotImplementedError
 
@@ -492,8 +492,11 @@ class BatchedFactory(Factory, ABC):
     def run(
         self,
         qp: QPROGRAM,
-        executor: Union[Callable[..., float], Callable[..., List[float]]],
-        scale_noise: Callable[[QPROGRAM, float], QPROGRAM],
+        executor: Executor | Callable[..., QuantumResult],
+        observable: Observable | None = None,
+        scale_noise: Callable[
+            [QPROGRAM, float], QPROGRAM
+        ] = fold_gates_at_random,  # type: ignore [has-type]
         num_to_average: int = 1,
     ) -> "BatchedFactory":
         """Computes the expectation values at each scale factor and stores them
@@ -512,15 +515,13 @@ class BatchedFactory(Factory, ABC):
 
         Args:
             qp: Quantum circuit to run.
-            executor: A "single executor" (1) or a "batched executor" (2).
-                (1) A function which inputs a single circuit and outputs a
-                single expectation value of interest.
-                (2) A function which inputs a list of circuits and outputs a
-                list of expectation values (one for each circuit). A batched
-                executor can also take an optional "kwargs_list" argument to
-                set a list of keyword arguments (one for each circuit). This
-                is necessary only if the factory is initialized using the
-                optional "shot_list" parameter.
+            executor: A ``mitiq.Executor`` or a function which inputs a (list
+                of) quantum circuits and outputs a (list of)
+                ``mitiq.QuantumResult`` s.
+            observable: Observable to compute the expectation value of. If
+                None, the `executor` must return an expectation value.
+                Otherwise, the `QuantumResult` returned by `executor` is used
+                to compute the expectation of the observable.
             scale_noise: Noise scaling function.
             num_to_average: The number of circuits executed for each noise
                 scale factor. This parameter can be used to increase the
@@ -532,28 +533,41 @@ class BatchedFactory(Factory, ABC):
 
         _check_circuit_length(qp)
 
-        # Get all noise-scaled circuits to run
+        # Get all noise-scaled circuits to run.
         to_run = self._generate_circuits(qp, scale_noise, num_to_average)
 
-        # Get the list of keywords associated to each circuit in "to_run"
+        # Run all circuits.
+        if not isinstance(executor, Executor):
+            executor = Executor(executor)
+
+        # Get the list of keywords associated to each circuit in "to_run".
         kwargs_list = self._get_keyword_args(num_to_average)
 
-        if Collector.is_batched_executor(executor):
-            if all([kwargs == {} for kwargs in kwargs_list]):
-                res = executor(to_run)
-            else:
-                res = executor(to_run, kwargs_list=kwargs_list)
+        # If there are different keyword args, run each circuit individually.
+        # https://stackoverflow.com/questions/1151658/python-hashable-dicts.
+        class HashableDict(dict[Any, Any]):
+            def __hash__(self) -> int:  # type: ignore[override]
+                return hash(tuple(sorted(self.items())))
+
+        if len(set(HashableDict(kwargs) for kwargs in kwargs_list)) != 1:
+            res = []
+            for circuit, kwargs in zip(to_run, kwargs_list):
+                res.extend(
+                    executor.evaluate(
+                        circuit, observable, force_run_all=True, **kwargs
+                    )
+                )
         else:
-            res = [
-                executor(circ, **kwargs)  # type: ignore
-                for circ, kwargs in zip(to_run, kwargs_list)
-            ]
+            # Else, run all circuits.
+            res = executor.evaluate(
+                to_run, observable, force_run_all=True, **kwargs_list[0]
+            )
 
         # Reshape "res" to have "num_to_average" columns
-        res = np.array(res).reshape((-1, num_to_average))
+        reshaped = np.array(res).reshape((-1, num_to_average))
 
         # Average the "num_to_average" columns
-        self._outstack = np.average(res, axis=1)
+        self._outstack = np.average(reshaped, axis=1).tolist()
 
         return self
 
@@ -584,7 +598,7 @@ class BatchedFactory(Factory, ABC):
         circuit: QPROGRAM,
         scale_noise: Callable[[QPROGRAM, float], QPROGRAM],
         num_to_average: int = 1,
-    ) -> List[QPROGRAM]:
+    ) -> list[QPROGRAM]:
         """Returns all noise-scaled circuits to run.
 
         Args:
@@ -611,7 +625,7 @@ class BatchedFactory(Factory, ABC):
                 {"scale_factor": scale} for scale in self._scale_factors
             ]
 
-    def _get_keyword_args(self, num_to_average: int) -> List[Dict[str, Any]]:
+    def _get_keyword_args(self, num_to_average: int) -> list[dict[str, Any]]:
         """Returns a list of keyword dictionaries to be used for
         executing the circuits generated by the method "_generate_circuits".
 
@@ -642,7 +656,7 @@ class AdaptiveFactory(Factory, ABC):
     """
 
     @abstractmethod
-    def next(self) -> Dict[str, float]:
+    def next(self) -> dict[str, float]:
         """Returns a dictionary of parameters to execute a circuit at."""
         raise NotImplementedError
 
@@ -705,8 +719,11 @@ class AdaptiveFactory(Factory, ABC):
     def run(
         self,
         qp: QPROGRAM,
-        executor: Callable[..., float],
-        scale_noise: Callable[[QPROGRAM, float], QPROGRAM],
+        executor: Executor | Callable[..., QuantumResult],
+        observable: Observable | None = None,
+        scale_noise: Callable[
+            [QPROGRAM, float], QPROGRAM
+        ] = fold_gates_at_random,  # type: ignore [has-type]
         num_to_average: int = 1,
         max_iterations: int = 100,
     ) -> "AdaptiveFactory":
@@ -716,9 +733,13 @@ class AdaptiveFactory(Factory, ABC):
 
         Args:
             qp: Circuit to mitigate.
-            executor: Function executing a circuit; returns an expectation
-                value. If shot_list is not None, then "shot" must be
-                an additional argument of the executor.
+            executor: A ``mitiq.Executor`` or a function which inputs a (list
+                of) quantum circuits and outputs a (list of)
+                ``mitiq.QuantumResult`` s.
+            observable: Observable to compute the expectation value of. If
+                None, the `executor` must return an expectation value.
+                Otherwise, the `QuantumResult` returned by `executor` is used
+                to compute the expectation of the observable.
             scale_noise: Function that scales the noise level of a quantum
                 circuit.
             num_to_average: Number of times expectation values are computed by
@@ -728,16 +749,21 @@ class AdaptiveFactory(Factory, ABC):
 
         _check_circuit_length(qp)
 
+        if not isinstance(executor, Executor):
+            executor = Executor(executor)
+
         def scale_factor_to_expectation_value(
             scale_factor: float, **exec_params: Any
         ) -> float:
             """Evaluates the quantum expectation value for a given
             scale_factor and other executor parameters."""
-            expectation_values = []
-            for _ in range(num_to_average):
-                scaled_qp = scale_noise(qp, scale_factor)
-                expectation_values.append(executor(scaled_qp, **exec_params))
-            return np.average(expectation_values)
+            to_run = [
+                scale_noise(qp, scale_factor) for _ in range(num_to_average)
+            ]
+            expectation_values = executor.evaluate(  # type: ignore[union-attr]
+                to_run, observable, force_run_all=True, **exec_params
+            )
+            return cast(float, np.average(expectation_values))
 
         return self.run_classical(
             scale_factor_to_expectation_value, max_iterations
@@ -745,8 +771,8 @@ class AdaptiveFactory(Factory, ABC):
 
 
 class PolyFactory(BatchedFactory):
-    """Factory object implementing a zero-noise extrapolation algorithm based on
-    a polynomial fit.
+    """Factory object implementing a zero-noise extrapolation algorithm based
+    on a polynomial fit.
 
     Args:
         scale_factors: Sequence of noise scale factors at which
@@ -771,9 +797,8 @@ class PolyFactory(BatchedFactory):
         self,
         scale_factors: Sequence[float],
         order: int,
-        shot_list: Optional[List[int]] = None,
+        shot_list: list[int] | None = None,
     ) -> None:
-        """Instantiates a new object of this Factory class."""
         if order > len(scale_factors) - 1:
             raise ValueError(
                 "The extrapolation order cannot exceed len(scale_factors) - 1."
@@ -832,7 +857,7 @@ class PolyFactory(BatchedFactory):
                 zne_error = np.sqrt(params_cov[order, order])
 
         def zne_curve(scale_factor: float) -> float:
-            return np.polyval(opt_params, scale_factor)
+            return cast(float, np.polyval(opt_params, scale_factor))
 
         return zne_limit, zne_error, opt_params, params_cov, zne_curve
 
@@ -928,7 +953,6 @@ class FakeNodesFactory(BatchedFactory):
         exp_values: Sequence[float],
         full_output: bool = False,
     ) -> ExtrapolationResult:
-
         if not FakeNodesFactory._is_equally_spaced(scale_factors):
             raise ValueError("The scale factors must be equally spaced.")
 
@@ -948,21 +972,23 @@ class FakeNodesFactory(BatchedFactory):
             opt_params,
             params_cov,
             zne_curve,
-        ) = RichardsonFactory.extrapolate(fake_nodes, exp_values, True)
+        ) = RichardsonFactory.extrapolate(  # type: ignore[misc]
+            fake_nodes, exp_values, True
+        )
 
         # Convert zne_curve from the "fake node space" to the real space.
         # Note: since a=0.0, this conversion is not necessary for zne_limit.
         def new_curve(scale_factor: float) -> float:
-            """Get real zne_cruve from the curve based on fake nodes."""
+            """Get real zne_curve from the curve based on fake nodes."""
             return zne_curve(
-                FakeNodesFactory._map_to_fake_nodes(scale_factor, a, b)
+                FakeNodesFactory._map_to_fake_nodes([scale_factor], a, b)[0]
             )
 
-        return zne_limit, zne_error, opt_params, params_cov, zne_curve
+        return zne_limit, zne_error, opt_params, params_cov, new_curve
 
     @staticmethod
     def _map_to_fake_nodes(
-        x: Union[Sequence[float], float], a: float, b: float
+        x: Sequence[float], a: float, b: float
     ) -> Sequence[float]:
         """
         A function that maps inputs to Chebyshev-Lobatto points. Based on
@@ -972,11 +998,9 @@ class FakeNodesFactory(BatchedFactory):
         we are mapping to.
 
         Args:
-            x:
-                Sequence[float]: Set of values to be mapped to CL points.
-                float: A single value to be mapped to a CL point.
-            a: A float representing the interval starting at a
-            b: A float representing the interval ending at b
+            x: Sequence[float]: Set of values to be mapped to CL points.
+            a: A float representing the interval starting at a.
+            b: A float representing the interval ending at b.
         Returns:
             A new sequence of fake nodes (Chebyshev-Lobatto points).
 
@@ -988,15 +1012,12 @@ class FakeNodesFactory(BatchedFactory):
         """
 
         # The mapping function
-        def S(x):
-            return (a - b) / 2 * np.cos(np.pi * (x - a) / (b - a)) + (
+        def mapping(_x: float) -> float:
+            return (a - b) / 2 * np.cos(np.pi * (_x - a) / (b - a)) + (
                 a + b
             ) / 2
 
-        if isinstance(x, float):
-            return S(x)
-
-        return np.array([S(y) for y in x])
+        return [mapping(y) for y in x]
 
     @staticmethod
     def _is_equally_spaced(arr: Sequence[float]) -> bool:
@@ -1023,10 +1044,6 @@ class LinearFactory(BatchedFactory):
     Raises:
         ValueError: If data is not consistent with the extrapolation model.
         ExtrapolationWarning: If the extrapolation fit is ill-conditioned.
-
-    Example:
-        >>> NOISE_LEVELS = [1.0, 2.0, 3.0]
-        >>> fac = LinearFactory(NOISE_LEVELS)
     """
 
     @staticmethod
@@ -1099,11 +1116,10 @@ class ExpFactory(BatchedFactory):
     def __init__(
         self,
         scale_factors: Sequence[float],
-        asymptote: Optional[float] = None,
+        asymptote: float | None = None,
         avoid_log: bool = False,
-        shot_list: Optional[List[int]] = None,
+        shot_list: list[int] | None = None,
     ) -> None:
-        """Instantiate an new object of this Factory class."""
         super(ExpFactory, self).__init__(scale_factors, shot_list)
         if not (asymptote is None or isinstance(asymptote, float)):
             raise ValueError(
@@ -1118,7 +1134,7 @@ class ExpFactory(BatchedFactory):
     def extrapolate(
         scale_factors: Sequence[float],
         exp_values: Sequence[float],
-        asymptote: Optional[float] = None,
+        asymptote: float | None = None,
         avoid_log: bool = False,
         eps: float = 1.0e-6,
         full_output: bool = False,
@@ -1219,11 +1235,10 @@ class PolyExpFactory(BatchedFactory):
         self,
         scale_factors: Sequence[float],
         order: int,
-        asymptote: Optional[float] = None,
+        asymptote: float | None = None,
         avoid_log: bool = False,
-        shot_list: Optional[List[int]] = None,
+        shot_list: list[int] | None = None,
     ) -> None:
-        """Instantiates a new object of this Factory class."""
         super(PolyExpFactory, self).__init__(scale_factors, shot_list)
         if not (asymptote is None or isinstance(asymptote, float)):
             raise ValueError(
@@ -1240,7 +1255,7 @@ class PolyExpFactory(BatchedFactory):
         scale_factors: Sequence[float],
         exp_values: Sequence[float],
         order: int,
-        asymptote: Optional[float] = None,
+        asymptote: float | None = None,
         avoid_log: bool = False,
         eps: float = 1.0e-6,
         full_output: bool = False,
@@ -1313,6 +1328,12 @@ class PolyExpFactory(BatchedFactory):
                 "The order cannot exceed the number"
                 f" of data points minus {1 + shift}."
             )
+        if not np.allclose(np.real(exp_values), exp_values):
+            raise ValueError(
+                f"Cannot extrapolate: Some expectation values in {exp_values} "
+                f"have non-zero imaginary part."
+            )
+        exp_values = np.real(exp_values).tolist()
 
         # Initialize default errors
         zne_error = None
@@ -1320,7 +1341,11 @@ class PolyExpFactory(BatchedFactory):
 
         # Deduce "sign" parameter of the exponential ansatz
         linear_params, _ = mitiq_polyfit(scale_factors, exp_values, deg=1)
-        sign = np.sign(-linear_params[0])
+
+        if asymptote is not None:
+            sign = np.sign(-(asymptote - linear_params[1]))
+        else:
+            sign = np.sign(-linear_params[0])
 
         def _ansatz_unknown(x: float, *coeffs: float) -> float:
             """Ansatz of generic order with unknown asymptote."""
@@ -1401,7 +1426,7 @@ class PolyExpFactory(BatchedFactory):
         # CASE 3: asymptote is given and "avoid_log" is False
         # Polynomial fit of z(x).
         shifted_y = [max(sign * (y - asymptote), eps) for y in exp_values]
-        zstack = np.log(shifted_y)
+        zstack = list(np.log(shifted_y))  # type: ignore
 
         # Get coefficients {z_j} of z(x)= z_0 + z_1*x + z_2*x**2...
         # Note: coefficients are ordered from high powers to powers of x
@@ -1438,8 +1463,8 @@ class PolyExpFactory(BatchedFactory):
 
 # Keep a log of the optimization process storing:
 # noise value(s), expectation value(s), parameters, and zero limit
-OptimizationHistory = List[
-    Tuple[List[Dict[str, float]], List[float], List[float], float]
+OptimizationHistory = list[
+    tuple[list[dict[str, float]], list[float], list[float], float]
 ]
 
 
@@ -1481,11 +1506,10 @@ class AdaExpFactory(AdaptiveFactory):
         self,
         steps: int,
         scale_factor: float = 2.0,
-        asymptote: Optional[float] = None,
+        asymptote: float | None = None,
         avoid_log: bool = False,
         max_scale_factor: float = 6.0,
     ) -> None:
-        """Instantiate a new object of this Factory class."""
         super(AdaExpFactory, self).__init__()
         if not (asymptote is None or isinstance(asymptote, float)):
             raise ValueError(
@@ -1514,7 +1538,7 @@ class AdaExpFactory(AdaptiveFactory):
         self.max_scale_factor = max_scale_factor
         self.history: OptimizationHistory = []
 
-    def next(self) -> Dict[str, float]:
+    def next(self) -> dict[str, float]:
         """Returns a dictionary of parameters to execute a circuit at."""
         # The 1st scale factor is always 1
         if len(self._instack) == 0:
@@ -1561,7 +1585,7 @@ class AdaExpFactory(AdaptiveFactory):
     def extrapolate(
         scale_factors: Sequence[float],
         exp_values: Sequence[float],
-        asymptote: Optional[float] = None,
+        asymptote: float | None = None,
         avoid_log: bool = False,
         eps: float = 1.0e-6,
         full_output: bool = False,
@@ -1629,7 +1653,7 @@ class AdaExpFactory(AdaptiveFactory):
             self._opt_params,
             self._params_cov,
             self._zne_curve,
-        ) = self.extrapolate(  # type: ignore
+        ) = self.extrapolate(  # type: ignore [misc]
             self.get_scale_factors(),
             self.get_expectation_values(),
             asymptote=self.asymptote,
