@@ -1,34 +1,23 @@
-# Copyright (C) 2020 Unitary Fund
+# Copyright (C) Unitary Foundation
 #
-# This program is free software: you can redistribute it and/or modify
-# it under the terms of the GNU General Public License as published by
-# the Free Software Foundation, either version 3 of the License, or
-# (at your option) any later version.
-#
-# This program is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-# GNU General Public License for more details.
-#
-# You should have received a copy of the GNU General Public License
-# along with this program.  If not, see <https://www.gnu.org/licenses/>.
+# This source code is licensed under the GPL license (v3) found in the
+# LICENSE file in the root directory of this source tree.
 
 """Functions for local and global unitary folding on supported circuits."""
+
+import warnings
 from copy import deepcopy
-from typing import (
-    Any,
-    Dict,
-    FrozenSet,
-    List,
-    Optional,
-    Tuple,
-)
+from typing import Any, cast
 
 import numpy as np
-from cirq import Circuit, InsertStrategy, inverse, ops, has_unitary
+from cirq import Circuit, InsertStrategy, Moment, has_unitary, inverse, ops
 
-from mitiq._typing import QPROGRAM
-from mitiq.interface import noise_scaling_converter
+from mitiq.interface import accept_qprogram_and_validate
+from mitiq.utils import (
+    _append_measurements,
+    _is_measurement,
+    _pop_measurements,
+)
 
 
 class UnfoldableCircuitError(Exception):
@@ -42,55 +31,31 @@ _cirq_gates_to_string_keys = {
     ops.Z: "Z",
     ops.T: "T",
     ops.I: "I",
+    ops.S: "S",
+    ops.T: "T",
+    ops.rx: "rx",
+    ops.ry: "ry",
+    ops.rz: "rz",
     ops.CNOT: "CNOT",
     ops.CZ: "CZ",
+    ops.SWAP: "SWAP",
+    ops.ISWAP: "ISWAP",
+    ops.CSWAP: "CSWAP",
     ops.TOFFOLI: "TOFFOLI",
 }
-_string_keys_to_cirq_gates = dict(
-    zip(_cirq_gates_to_string_keys.values(), _cirq_gates_to_string_keys.keys())
-)
+_string_keys_to_cirq_gates = {
+    opstring: op for op, opstring in _cirq_gates_to_string_keys.items()
+}
+
+_valid_gate_names = list(
+    map(
+        lambda gate_name: gate_name.lower(),
+        _cirq_gates_to_string_keys.values(),
+    )
+) + ["single", "double", "triple"]
 
 
 # Helper functions
-def _is_measurement(op: ops.Operation) -> bool:
-    """Returns true if the operation's gate is a measurement, else False.
-
-    Args:
-        op: Gate operation.
-    """
-    return isinstance(op.gate, ops.measurement_gate.MeasurementGate)
-
-
-def _pop_measurements(circuit: Circuit,) -> List[Tuple[int, ops.Operation]]:
-    """Removes all measurements from a circuit.
-
-    Args:
-        circuit: A quantum circuit as a :class:`cirq.Circuit` object.
-
-    Returns:
-        measurements: List of measurements in the circuit.
-    """
-    measurements = list(circuit.findall_operations(_is_measurement))
-    circuit.batch_remove(measurements)
-    return measurements
-
-
-def _append_measurements(
-    circuit: Circuit, measurements: List[Tuple[int, ops.Operation]]
-) -> None:
-    """Appends all measurements into the final moment of the circuit.
-
-    Args:
-        circuit: a quantum circuit as a :class:`cirq.Circuit`.
-        measurements: measurements to perform.
-    """
-    new_measurements: List[Tuple[int, ops.Operation]] = []
-    for i in range(len(measurements)):
-        # Make sure the moment to insert into is the last in the circuit
-        new_measurements.append((len(circuit) + 1, measurements[i][1]))
-    circuit.batch_insert(new_measurements)
-
-
 def _check_foldable(circuit: Circuit) -> None:
     """Raises an error if the input circuit cannot be folded.
 
@@ -109,10 +74,14 @@ def _check_foldable(circuit: Circuit) -> None:
         )
 
     if not has_unitary(circuit):
-        raise UnfoldableCircuitError(
-            "Circuit contains non-unitary channels which are not terminal "
-            "measurements and cannot be folded."
-        )
+        circ_measurements = _pop_measurements(circuit)
+        if inverse(circuit, default=None) is None:
+            raise UnfoldableCircuitError(
+                "Circuit contains non-invertible channels which are not"
+                "terminal measurements and cannot be folded."
+            )
+        else:
+            _append_measurements(circuit, circ_measurements)
 
 
 def _squash_moments(circuit: Circuit) -> Circuit:
@@ -122,18 +91,19 @@ def _squash_moments(circuit: Circuit) -> Circuit:
     Args:
         circuit: Circuit to squash moments of.
     """
-    return Circuit(
-        circuit.all_operations(),
-        strategy=InsertStrategy.EARLIEST,
-        device=circuit.device,
+    output_circuit = circuit.copy()
+    output_circuit = output_circuit[0:0]  # Remove moments
+    output_circuit.append(
+        circuit.all_operations(), strategy=InsertStrategy.EARLIEST
     )
+    return output_circuit
 
 
 def _fold_all(
     circuit: Circuit,
     num_folds: int = 1,
-    exclude: FrozenSet[Any] = frozenset(),
-    skip_moments: FrozenSet[int] = frozenset(),
+    exclude: frozenset[Any] = frozenset(),
+    skip_moments: frozenset[int] = frozenset(),
 ) -> Circuit:
     """Returns a circuit with all gates folded locally.
 
@@ -151,7 +121,7 @@ def _fold_all(
         )
 
     # Parse the exclude argument.
-    all_gates = set(op.gate for op in circuit.all_operations())
+    all_gates = set(cast(ops.Gate, op.gate) for op in circuit.all_operations())
     to_exclude = set()
     for item in exclude:
         if isinstance(item, str):
@@ -184,7 +154,7 @@ def _fold_all(
             )
 
     folded = deepcopy(circuit)[:0]
-    for (i, moment) in enumerate(circuit):
+    for i, moment in enumerate(circuit):
         if i in skip_moments:
             folded.append(moment, strategy=InsertStrategy.EARLIEST)
             continue
@@ -207,7 +177,7 @@ def _default_weight(op: ops.Operation) -> float:
 
 
 def _get_weight_for_gate(
-    weights: Optional[Dict[str, float]], op: ops.Operation
+    weights: dict[str, float], op: ops.Operation
 ) -> float:
     """Returns the weight for a given gate, using a default value of 1.0 if
     weights is None or if the weight is not specified.
@@ -236,12 +206,12 @@ def _get_weight_for_gate(
 
 
 # Local folding functions
-@noise_scaling_converter
+@accept_qprogram_and_validate
 def fold_all(
-    circuit: QPROGRAM,
+    circuit: Circuit,
     scale_factor: float,
-    exclude: FrozenSet[Any] = frozenset(),
-) -> QPROGRAM:
+    exclude: frozenset[Any] = frozenset(),
+) -> Circuit:
     """Returns a circuit with all gates folded locally.
 
     Args:
@@ -266,8 +236,16 @@ def fold_all(
                 "Y"         | Pauli Y
                 "Z"         | Pauli Z
                 "I"         | Identity
+                "S"         | Phase gate
+                "T"         | T gate
+                "rx"        | X-rotation
+                "ry"        | Y-rotation
+                "rz"        | Z-rotation
                 "CNOT"      | CNOT
                 "CZ"        | CZ gate
+                "SWAP"      | Swap
+                "ISWAP"     | Imaginary swap
+                "CSWAP"     | CSWAP
                 "TOFFOLI"   | Toffoli gate
                 "single"    | All single qubit gates
                 "double"    | All two-qubit gates
@@ -289,10 +267,10 @@ def fold_all(
 
 
 # Global folding function
-@noise_scaling_converter
+@accept_qprogram_and_validate
 def fold_global(
-    circuit: QPROGRAM, scale_factor: float, **kwargs: Any
-) -> QPROGRAM:
+    circuit: Circuit, scale_factor: float, **kwargs: Any
+) -> Circuit:
     """Returns a new circuit obtained by folding the global unitary of the
     input circuit.
 
@@ -304,10 +282,6 @@ def fold_global(
         scale_factor: Factor to scale the circuit by.
 
     Keyword Args:
-        squash_moments (bool): If True, all gates (including folded gates) are
-            placed as early as possible in the circuit. If False, new moments
-            are created for folded gates. This option only applies to QPROGRAM
-            types which have a "moment" or "time" structure. Default is True.
         return_mitiq (bool): If True, returns a Mitiq circuit instead of
             the input circuit type (if different). Default is False.
 
@@ -334,20 +308,31 @@ def fold_global(
     num_to_fold = int(round(fraction_scale * len(operations) / 2))
 
     if num_to_fold > 0:
-        folded += Circuit(
-            [inverse(operations[-num_to_fold:])], [operations[-num_to_fold:]]
-        )
+        # Create the inverse of the final partial circuit
+        inverse_partial = Circuit()
+        num_partial = 0
+        for moment in base_circuit[::-1]:
+            new_moment = Moment()
+            for op in moment.operations[::-1]:
+                new_moment = new_moment.with_operation(inverse(op))
+                num_partial += 1
+                if num_partial == num_to_fold:
+                    break
+            inverse_partial.append(new_moment)
+            if num_partial == num_to_fold:
+                break
+        # Append partially folded circuit
+        folded += inverse_partial + inverse(inverse_partial)
 
     _append_measurements(folded, measurements)
-    if not (kwargs.get("squash_moments") is False):
-        folded = _squash_moments(folded)
     return folded
 
 
 def _create_weight_mask(
-    circuit: Circuit, fidelities: Optional[Dict[str, float]],
-) -> List[float]:
-    """Returns a list of weights associated to each gate if the input
+    circuit: Circuit,
+    fidelities: dict[str, float],
+) -> list[float]:
+    """Returns a list of weights associated to each gate in the input
     circuit. Measurement gates are ignored.
 
     The gate ordering is equal to the one used in the `all_operations()`
@@ -363,12 +348,20 @@ def _create_weight_mask(
 
     Returns: The list of weights associated to all the gates.
     """
-    if fidelities and not all(0.0 < f <= 1.0 for f in fidelities.values()):
+    if not all(0.0 < f <= 1.0 for f in fidelities.values()):
         raise ValueError("Fidelities should be in the interval (0, 1].")
-    weights = None
-    if fidelities:
-        # Round to avoid ugly numbers like 0.09999999999999998 instead of 0.1
-        weights = {k: round(1.0 - f, 12) for k, f in fidelities.items()}
+
+    invalid_fidelities = filter(
+        lambda opname: opname.lower() not in _valid_gate_names, fidelities
+    )
+    for gate_name in invalid_fidelities:
+        warnings.warn(
+            f"You passed a fidelity for the gate '{gate_name}', but we don't "
+            "currently support this gate."
+        )
+
+    # Round to avoid ugly numbers like 0.09999999999999998 instead of 0.1
+    weights = {k: round(1.0 - f, 12) for k, f in fidelities.items()}
 
     # Build mask with weights of each gate
     return [
@@ -379,11 +372,10 @@ def _create_weight_mask(
 
 
 def _create_fold_mask(
-    weight_mask: List[float],
+    weight_mask: list[float],
     scale_factor: float,
-    folding_method: str,
-    seed: Optional[int] = None,
-) -> List[float]:
+    seed: int | None = None,
+) -> list[int]:
     r"""Returns a list of integers determining how many times each gate a
     circuit should be folded to realize the desired input scale_factor.
 
@@ -401,19 +393,16 @@ def _create_fold_mask(
     total weight of the input circuit.
 
     For equal weights, this function reproduces the local unitary folding
-    method defined in equation (5) of :cite:`Giurgica_Tiron_2020_arXiv`.
+    method defined in equation (5) of :cite:`Giurgica_Tiron_2020_arXiv`, with
+    the layer indices chosen at random.
 
     Args:
         weight_mask: The weights of all the gates of the circuit to fold.
             Highly noisy gates should have a corresponding high weight.
             Gates with zero weight are assumed to be ideal and are not folded.
         scale_factor: The effective noise scale factor.
-        folding_method: A string equal to "at_random", or "from_left", or
-            "from_right". Determines the partial folding method described in
-            :cite:`Giurgica_Tiron_2020_arXiv`. If scale_factor is an odd
-            integer, all methods are equivalent and this option is irrelevant.
-        seed: A seed for the random number generator. This is used only when
-            folding_method is "at_random".
+        seed: A seed for the random number generator used to select the
+            subset of layer indices to fold.
 
     Returns: The list of integers determining to how many times one should
         fold the j_th gate of the circuit to be scaled.
@@ -422,7 +411,6 @@ def _create_fold_mask(
         >>>_create_fold_mask(
             weight_mask=[1.0, 0.5, 2.0, 0.0],
             scale_factor=4,
-            folding_method="from_left",
         )
         [2, 2, 1, 0]
     """
@@ -450,18 +438,9 @@ def _create_fold_mask(
 
     # Express folding order through a list of indices
     folding_order = list(range(len(weight_mask)))
-    if folding_method == "from_left":
-        pass
-    elif folding_method == "from_right":
-        folding_order.reverse()
-    elif folding_method == "at_random":
-        rnd_state = np.random.RandomState(seed)
-        rnd_state.shuffle(folding_order)
-    else:
-        raise ValueError(
-            "The option 'folding_method' is not valid."
-            "It must be 'at_random', or 'from_left', or 'from_right'."
-        )
+
+    rnd_state = np.random.RandomState(seed)
+    rnd_state.shuffle(folding_order)
 
     # Fold gates until the input scale_factor is better approximated
     input_circuit_weight = sum(weight_mask)
@@ -490,8 +469,8 @@ def _create_fold_mask(
 
 def _apply_fold_mask(
     circuit: Circuit,
-    num_folds_mask: List[int],
-    squash_moments: Optional[bool] = True,
+    num_folds_mask: list[int],
+    squash_moments: bool = True,
 ) -> Circuit:
     r"""Applies local unitary folding to the gates of the input circuit
     according to the input num_folds_mask.
@@ -547,170 +526,17 @@ def _apply_fold_mask(
     return folded_circuit
 
 
-@noise_scaling_converter
-def fold_gates_from_left(
-    circuit: QPROGRAM, scale_factor: float, **kwargs: Any
-) -> QPROGRAM:
-    """Returns a new folded circuit by applying the map G -> G G^dag G to a
-    subset of gates of the input circuit, starting with gates at the
-    left (beginning) of the circuit.
-
-    The folded circuit has a number of gates approximately equal to
-    scale_factor * n where n is the number of gates in the input circuit.
-
-    For equal gate fidelities, this function reproduces the local unitary
-    folding method defined in equation (5) of
-    :cite:`Giurgica_Tiron_2020_arXiv`.
-
-    Args:
-        circuit: Circuit to fold.
-        scale_factor: Factor to scale the circuit by. Any real number >= 1.
-
-    Keyword Args:
-        fidelities (Dict[str, float]): Dictionary of gate fidelities. Each key
-            is a string which specifies the gate and each value is the
-            fidelity of that gate. When this argument is provided, folded
-            gates contribute an amount proportional to their infidelity
-            (1 - fidelity) to the total noise scaling. Fidelity values must be
-            in the interval (0, 1]. Gates not specified have a default
-            fidelity of 0.99**n where n is the number of qubits the gates act
-            on.
-
-            Supported gate keys are listed in the following table.::
-
-                Gate key    | Gate
-                -------------------------
-                "H"         | Hadamard
-                "X"         | Pauli X
-                "Y"         | Pauli Y
-                "Z"         | Pauli Z
-                "I"         | Identity
-                "CNOT"      | CNOT
-                "CZ"        | CZ gate
-                "TOFFOLI"   | Toffoli gate
-                "single"    | All single qubit gates
-                "double"    | All two-qubit gates
-                "triple"    | All three-qubit gates
-
-            Keys for specific gates override values set by "single", "double",
-            and "triple".
-
-            For example, `fidelities = {"single": 1.0, "H", 0.99}` sets all
-            single-qubit gates except Hadamard to have fidelity one.
-
-        squash_moments (bool): If True, all gates (including folded gates) are
-            placed as early as possible in the circuit. If False, new moments
-            are created for folded gates. This option only applies to QPROGRAM
-            types which have a "moment" or "time" structure. Default is True.
-
-        return_mitiq (bool): If True, returns a Mitiq circuit instead of
-            the input circuit type (if different). Default is False.
-
-    Returns:
-        folded: The folded quantum circuit as a QPROGRAM.
-
-    """
-
-    weight_mask = _create_weight_mask(circuit, kwargs.get("fidelities"))
-
-    num_folds_mask = _create_fold_mask(
-        weight_mask, scale_factor, folding_method="from_left"
-    )
-
-    return _apply_fold_mask(
-        circuit,
-        num_folds_mask,
-        squash_moments=kwargs.get("squash_moments", True),
-    )
-
-
-@noise_scaling_converter
-def fold_gates_from_right(
-    circuit: QPROGRAM, scale_factor: float, **kwargs: Any
-) -> QPROGRAM:
-    r"""Returns a new folded circuit by applying the map G -> G G^dag G to a
-    subset of gates of the input circuit, starting with gates at the
-    right (end) of the circuit.
-
-    The folded circuit has a number of gates approximately equal to
-    scale_factor * n where n is the number of gates in the input circuit.
-
-    For equal gate fidelities, this function reproduces the local unitary
-    folding method defined in equation (5) of
-    :cite:`Giurgica_Tiron_2020_arXiv`.
-
-    Args:
-        circuit: Circuit to fold.
-        scale_factor: Factor to scale the circuit by. Any real number >= 1.
-
-    Keyword Args:
-        fidelities (Dict[str, float]): Dictionary of gate fidelities. Each key
-            is a string which specifies the gate and each value is the
-            fidelity of that gate. When this argument is provided, folded
-            gates contribute an amount proportional to their infidelity
-            (1 - fidelity) to the total noise scaling. Fidelity values must be
-            in the interval (0, 1]. Gates not specified have a default
-            fidelity of 0.99**n where n is the number of qubits the gates act
-            on.
-
-            Supported gate keys are listed in the following table.::
-
-                Gate key    | Gate
-                -------------------------
-                "H"         | Hadamard
-                "X"         | Pauli X
-                "Y"         | Pauli Y
-                "Z"         | Pauli Z
-                "I"         | Identity
-                "CNOT"      | CNOT
-                "CZ"        | CZ gate
-                "TOFFOLI"   | Toffoli gate
-                "single"    | All single qubit gates
-                "double"    | All two-qubit gates
-                "triple"    | All three-qubit gates
-
-            Keys for specific gates override values set by "single", "double",
-            and "triple".
-
-            For example, `fidelities = {"single": 1.0, "H", 0.99}` sets all
-            single-qubit gates except Hadamard to have fidelity one.
-
-        squash_moments (bool): If True, all gates (including folded gates) are
-            placed as early as possible in the circuit. If False, new moments
-            are created for folded gates. This option only applies to QPROGRAM
-            types which have a "moment" or "time" structure. Default is True.
-
-        return_mitiq (bool): If True, returns a Mitiq circuit instead of
-            the input circuit type (if different). Default is False.
-
-    Returns:
-        folded: The folded quantum circuit as a QPROGRAM.
-
-    """
-
-    weight_mask = _create_weight_mask(circuit, kwargs.get("fidelities"))
-
-    num_folds_mask = _create_fold_mask(
-        weight_mask, scale_factor, folding_method="from_right"
-    )
-
-    return _apply_fold_mask(
-        circuit,
-        num_folds_mask,
-        squash_moments=kwargs.get("squash_moments", True),
-    )
-
-
-@noise_scaling_converter
+@accept_qprogram_and_validate
 def fold_gates_at_random(
-    circuit: QPROGRAM,
+    circuit: Circuit,
     scale_factor: float,
-    seed: Optional[int] = None,
+    seed: int | None = None,
     **kwargs: Any,
-) -> QPROGRAM:
-    r"""Returns a new folded circuit by applying the map G -> G G^dag G to a
-    subset of gates of the input circuit, starting with gates at the
-    right (end) of the circuit.
+) -> Circuit:
+    r"""
+    Returns a new folded circuit by applying the map G -> G G^dag G to a
+    subset of gates of the input circuit, different indices randomly sampled
+    without replacement.
 
     The folded circuit has a number of gates approximately equal to
     scale_factor * n where n is the number of gates in the input circuit.
@@ -722,6 +548,7 @@ def fold_gates_at_random(
     Args:
         circuit: Circuit to fold.
         scale_factor: Factor to scale the circuit by. Any real number >= 1.
+        seed: Seed for random number generator.
 
     Keyword Args:
         fidelities (Dict[str, float]): Dictionary of gate fidelities. Each key
@@ -742,8 +569,16 @@ def fold_gates_at_random(
                 "Y"         | Pauli Y
                 "Z"         | Pauli Z
                 "I"         | Identity
+                "S"         | Phase gate
+                "T"         | T gate
+                "rx"        | X-rotation
+                "ry"        | Y-rotation
+                "rz"        | Z-rotation
                 "CNOT"      | CNOT
                 "CZ"        | CZ gate
+                "SWAP"      | Swap
+                "ISWAP"     | Imaginary swap
+                "CSWAP"     | CSWAP
                 "TOFFOLI"   | Toffoli gate
                 "single"    | All single qubit gates
                 "double"    | All two-qubit gates
@@ -768,10 +603,12 @@ def fold_gates_at_random(
 
     """
 
-    weight_mask = _create_weight_mask(circuit, kwargs.get("fidelities"))
+    weight_mask = _create_weight_mask(circuit, kwargs.get("fidelities", {}))
 
     num_folds_mask = _create_fold_mask(
-        weight_mask, scale_factor, folding_method="at_random", seed=seed,
+        weight_mask,
+        scale_factor,
+        seed=seed,
     )
 
     return _apply_fold_mask(
